%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(process_SUITE).

%% Tests processes, trapping exit messages and the BIFs:
%% 	exit/1
%%	exit/2
%%	process_info/1,2

-include("test_server.hrl").

-define(heap_binary_size, 64).

-export([all/1, spawn_with_binaries/1,
	 t_exit_1/1, t_exit_2/1, t_exit_2_other/1, t_exit_2_other_normal/1,
	 self_exit/1, normal_suicide_exit/1, abnormal_suicide_exit/1,
	 t_exit_2_catch/1, trap_exit_badarg/1, trap_exit_badarg_in_bif/1,
	 exit_and_timeout/1, exit_twice/1, t_process_info/1,
	 bump_reductions/1, low_prio/1, binary_owner/1, yield/1,
	 otp_4725/1]).
-export([prio_server/2, prio_client/2]).

-export([init_per_testcase/2, fin_per_testcase/2]).

all(suite) -> [spawn_with_binaries, t_exit_1, t_exit_2,
	       trap_exit_badarg, trap_exit_badarg_in_bif,
	       t_process_info, bump_reductions,
	       low_prio, yield, otp_4725].

init_per_testcase(Func, Config) when atom(Func), list(Config) ->
    Dog=?t:timetrap(?t:minutes(10)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Func, Config) ->
    Dog=?config(watchdog, Config),
    ?t:timetrap_cancel(Dog).

fun_spawn(Fun) ->
    spawn_link(erlang, apply, [Fun, []]).

spawn_with_binaries(doc) ->
    "Tests that binaries as arguments to spawn/3 doesn't leak "
	"(expect \"Got unsuable memory block\" if there is a leak).";
spawn_with_binaries(suite) -> [];
spawn_with_binaries(Config) when list(Config) ->
    ?line L = lists:duplicate(2048, 42),
    ?line TwoMeg = lists:duplicate(1024, L),
    ?line Fun = fun() -> spawn(?MODULE, binary_owner, [list_to_binary(TwoMeg)]),
			 receive after 1 -> ok end end,
    ?line Iter = case test_server:purify_is_running() of
		     true -> 10;
		     false -> 150
		 end,
    ?line test_server:do_times(Iter, Fun),
    ok.

binary_owner(Bin) ->
    ok.

t_exit_1(doc) -> "Tests exit/1 with a big message.";
t_exit_1(suite) -> [];
t_exit_1(Config) when list(Config) ->
    ?line start_spawner(),
    ?line Dog = test_server:timetrap(test_server:seconds(20)),
    ?line process_flag(trap_exit, true),
    ?line test_server:do_times(10, fun t_exit_1/0),
    ?line test_server:timetrap_cancel(Dog),
    ?line stop_spawner(),
    ok.

t_exit_1() ->
    ?line Pid = fun_spawn(fun() -> exit(kb_128()) end),
    ?line Garbage = kb_128(),
    ?line receive
	      {'EXIT', Pid, Garbage} -> ok
	  end.

t_exit_2(suite) -> [t_exit_2_other, t_exit_2_other_normal, 
		    self_exit, normal_suicide_exit,
		    abnormal_suicide_exit, t_exit_2_catch,
		    exit_and_timeout, exit_twice].

t_exit_2_other(doc) -> "Tests exit/2 with a lot of data in the exit message.";
t_exit_2_other(suite) -> [];
t_exit_2_other(Config) when list(Config) ->
    ?line start_spawner(),
    ?line Dog = test_server:timetrap(test_server:seconds(20)),
    ?line process_flag(trap_exit, true),
    ?line test_server:do_times(10, fun t_exit_2_other/0),
    ?line test_server:timetrap_cancel(Dog),
    ?line stop_spawner(),
    ok.

t_exit_2_other() ->
    ?line Pid = fun_spawn(fun() -> receive x -> ok end end),
    ?line Garbage = kb_128(),
    ?line exit(Pid, Garbage),
    ?line receive
	      {'EXIT', Pid, Garbage} -> ok
	  end.

t_exit_2_other_normal(suite) -> [];
t_exit_2_other_normal(doc) ->
    "Tests that exit(Pid, normal) does not kill another process.";
t_exit_2_other_normal(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(20)),
    ?line process_flag(trap_exit, true),
    ?line Pid = fun_spawn(fun() -> receive x -> ok end end),
    ?line exit(Pid, normal),
    ?line receive
	      {'EXIT', Pid, Reason} ->
		  ?line test_server:fail({process_died, Reason})
	  after 1000 ->
		  ok
	  end,
    ?line case process_info(Pid) of
	      undefined ->
		  test_server:fail(process_died_on_normal);
	      List when list(List) ->
		  ok
	  end,
    exit(Pid, kill),
    ?line test_server:timetrap_cancel(Dog),
    ok.

self_exit(doc) ->
    "Tests that we can trap an exit message sent with exit/2 from "
    "the same process.";
self_exit(suite) -> [];
self_exit(Config) when list(Config) ->
    ?line start_spawner(),
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line process_flag(trap_exit, true),
    ?line test_server:do_times(200, fun self_exit/0),
    ?line test_server:timetrap_cancel(Dog),
    ?line stop_spawner(),
    ok.

self_exit() ->
    ?line Garbage = eight_kb(),
    ?line P = self(),
    ?line true = exit(P, Garbage),
    ?line receive
	      {'EXIT', P, Garbage} -> ok
	  end.

normal_suicide_exit(doc) ->
    "Tests exit(self(), normal) is equivalent to exit(normal) for a process "
    "that doesn't trap exits.";
normal_suicide_exit(suite) -> [];
normal_suicide_exit(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line Pid = fun_spawn(fun() -> exit(self(), normal) end),
    ?line receive
	      {'EXIT', Pid, normal} -> ok;
	      Other -> test_server:fail({bad_message, Other})
	  end.

abnormal_suicide_exit(doc) ->
    "Tests exit(self(), Term) is equivalent to exit(Term) for a process "
    "that doesn't trap exits.";
abnormal_suicide_exit(suite) -> [];
abnormal_suicide_exit(Config) when list(Config) ->
    ?line Garbage = eight_kb(),
    ?line process_flag(trap_exit, true),
    ?line Pid = fun_spawn(fun() -> exit(self(), Garbage) end),
    ?line receive
	      {'EXIT', Pid, Garbage} -> ok;
	      Other -> test_server:fail({bad_message, Other})
	  end.

t_exit_2_catch(doc) ->
    "Tests that exit(self(), die) cannot be catched.";
t_exit_2_catch(suite) -> [];
t_exit_2_catch(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line Pid = fun_spawn(fun() -> catch exit(self(), die) end),
    ?line receive
	      {'EXIT', Pid, normal} ->
		  test_server:fail(catch_worked);
	      {'EXIT', Pid, die} ->
		  ok;
	      Other ->
		  test_server:fail({bad_message, Other})
	  end.

trap_exit_badarg(doc) ->
    "Tests trapping of an 'EXIT' message generated by a bad argument to "
    "the abs/1 bif.  The 'EXIT' message will intentionally be very big.";
trap_exit_badarg(suite) -> [];
trap_exit_badarg(Config) when list(Config) ->
    ?line start_spawner(),
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line process_flag(trap_exit, true),
    ?line test_server:do_times(10, fun trap_exit_badarg/0),
    ?line test_server:timetrap_cancel(Dog),
    ?line stop_spawner(),
    ok.

trap_exit_badarg() ->
    ?line Pid = fun_spawn(fun() -> bad_guy(kb_128()) end),
    ?line Garbage = kb_128(),
    ?line receive
	      {'EXIT', Pid, {badarg,[{erlang,abs,[Garbage]},{?MODULE,bad_guy,1}|_]}} ->
		  ok;
	      Other ->
		  ?line ok = io:format("Bad EXIT message: ~P", [Other, 30]),
		  ?line test_server:fail(bad_exit_message)
	  end.

bad_guy(Arg) ->
    ?line abs(Arg).


kb_128() ->
    Eight = eight_kb(),
    {big_binary(),
     Eight, Eight, Eight, Eight, Eight, Eight, Eight, Eight,
     big_binary(),
     Eight, Eight, Eight, Eight, Eight, Eight, Eight, Eight,
     big_binary()}.

eight_kb() ->
    %%% This is really much more than eight kb, so vxworks platforms
    %%% gets away with 1/8 of the other platforms (due to limited
    %%% memory resources). 
    B64 = case os:type() of
	      vxworks ->
		  ?line lists:seq(1, 8);
	      _ ->
		  ?line lists:seq(1, 64)
	  end,
    ?line B512 = {<<1>>,B64,<<2,3>>,B64,make_sub_binary([4,5,6,7,8,9]),
		  B64,make_sub_binary([1,2,3,4,5,6]),
		  B64,make_sub_binary(lists:seq(1, ?heap_binary_size+1)),
		  B64,B64,B64,B64,big_binary()},
    ?line lists:duplicate(8, {B512,B512}).

big_binary() ->
    big_binary(10, [42]).
big_binary(0, Acc) ->
    list_to_binary(Acc);
big_binary(N, Acc) ->
    big_binary(N-1, [Acc|Acc]).

trap_exit_badarg_in_bif(doc) ->
    "Test receiving an EXIT message when spawning a BIF with bad arguments.";
trap_exit_badarg_in_bif(suite) ->
    [];
trap_exit_badarg_in_bif(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line process_flag(trap_exit, true),
    ?line test_server:do_times(10, fun trap_exit_badarg_bif/0),
    ?line test_server:timetrap_cancel(Dog),
    ok.
    
trap_exit_badarg_bif() ->
    ?line Pid = spawn_link(erlang, node, [1]),
    ?line receive
	      {'EXIT', Pid, {badarg, _}} ->
		  ok;
	      Other ->
		  ?line test_server:fail({unexpected, Other})
	  end.

%% The following sequences of events have crasched Beam.
%%
%% 1) An exit is sent to a process which is currently not running.
%%    The exit reason will (on purpose) overwrite the message queue
%%    pointer.
%% 2) Before the process is scheduled in, it receives a timeout (from
%%    a 'receive after').
%% 3) The process will crash the next time it executes 'receive'.

exit_and_timeout(doc) ->
    "Test for a previous crash in Beam. Complete description in test case.";
exit_and_timeout(suite) -> [];
exit_and_timeout(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(20)),

    ?line process_flag(trap_exit, true),
    ?line Parent = self(),
    ?line Low = fun_spawn(fun() -> eat_low(Parent) end),
    ?line High = fun_spawn(fun() -> eat_high(Low) end),
    ?line eat_wait_for(Low, High),

    ?line test_server:timetrap_cancel(Dog),
    ok.


eat_wait_for(Low, High) ->
    ?line receive
	      {'EXIT', Low, {you, are, dead}} ->
		  ok;
	      {'EXIT', High, normal} ->
		  eat_wait_for(Low, High);
	      Other ->
		  test_server:fail({bad_message, Other})
	  end.

eat_low(Parent) ->
    receive
    after 2500 ->
	    ok
    end,
    receive
	Any ->
	    io:format("Received: ~p\n", [Any])
    after 1000 ->
	    ok
    end.
    
eat_high(Low) ->
    process_flag(priority, high),
    receive after 1000 -> ok end,
    exit(Low, {you, are, dead}),
    {_, Sec, _} = now(),
    loop(Sec, Sec).

%% Busy loop for 5 seconds.

loop(OrigSec, CurrentSec) when CurrentSec < OrigSec+5 ->
    {_, NewSec, _} = now(),
    loop(OrigSec, NewSec);
loop(_, _) ->
    ok.


exit_twice(doc) ->
    "Tries to send two different exit messages to a process. "
    "(The second one should be ignored.)";
exit_twice(suite) -> [];
exit_twice(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(20)),

    ?line process_flag(trap_exit, true),
    ?line Low = fun_spawn(fun etwice_low/0),
    ?line High = fun_spawn(fun() -> etwice_high(Low) end),
    ?line etwice_wait_for(Low, High),

    ?line test_server:timetrap_cancel(Dog),
    ok.

etwice_wait_for(Low, High) ->
    ?line receive
	      {'EXIT', Low, first} ->
		  ok;
	      {'EXIT', Low, Other} ->
		  test_server:fail({wrong_exit_reason, Other});
	      {'EXIT', High, normal} ->
		  etwice_wait_for(Low, High);
	      Other ->
		  test_server:fail({bad_message, Other})
	  end.

etwice_low() ->
    etwice_low().

etwice_high(Low) ->
    process_flag(priority, high),
    exit(Low, first),
    exit(Low, second).

t_process_info(doc) -> "Tests the process_info/1 BIF.";
t_process_info(suit) -> [];
t_process_info(Config) when list(Config) ->
    ?line [] = process_info(self(), registered_name),
    ?line register(my_name, self()),
    ?line {registered_name, my_name} = process_info(self(), registered_name),
    ?line {status, running} = process_info(self(), status),
    ?line {current_function, {?MODULE, t_process_info, 1}} =
	process_info(self(), current_function),
    ?line Gleader = group_leader(),
    ?line {group_leader, Gleader} = process_info(self(), group_leader),
    ?line {'EXIT',{badarg,Info}} = (catch process_info('not_a_pid')),
    ok.

otp_4725(doc) -> [];
otp_4725(suite) -> [];
otp_4725(Config) when list(Config) ->
    ?line Tester = self(),
    ?line Ref1 = make_ref(),
    ?line Pid1 = spawn_opt(fun () ->
				   Tester ! {Ref1, process_info(self())},
				   receive
				       Ref1 -> bye
				   end
			   end,
			   [link,
			    {priority, max},
			    {fullsweep_after, 600}]),
    ?line receive
	      {Ref1, ProcInfo1A} ->
		  ?line ProcInfo1B = process_info(Pid1),
		  ?line Pid1 ! Ref1,
		  ?line check_proc_infos(ProcInfo1A, ProcInfo1B)
	  end,
    ?line Ref2 = make_ref(),
    ?line Pid2 = spawn_opt(fun () ->
				   Tester ! {Ref2, process_info(self())},
				   receive
				       Ref2 -> bye
				   end
			   end,
			   []),
    ?line receive
	      {Ref2, ProcInfo2A} ->
		  ?line ProcInfo2B = process_info(Pid2),
		  ?line Pid2 ! Ref2,
		  ?line check_proc_infos(ProcInfo2A, ProcInfo2B)
	  end,
    ?line ok.

check_proc_infos(A, B) ->
    ?line IC = lists:keysearch(initial_call, 1, A),
    ?line IC = lists:keysearch(initial_call, 1, B),

    ?line L = lists:keysearch(links, 1, A),
    ?line L = lists:keysearch(links, 1, B),

    ?line D = lists:keysearch(dictionary, 1, A),
    ?line D = lists:keysearch(dictionary, 1, B),

    ?line TE = lists:keysearch(trap_exit, 1, A),
    ?line TE = lists:keysearch(trap_exit, 1, B),

    ?line EH = lists:keysearch(error_handler, 1, A),
    ?line EH = lists:keysearch(error_handler, 1, B),

    ?line P = lists:keysearch(priority, 1, A),
    ?line P = lists:keysearch(priority, 1, B),

    ?line GL = lists:keysearch(group_leader, 1, A),
    ?line GL = lists:keysearch(group_leader, 1, B),

    ?line GC = lists:keysearch(garbage_collection, 1, A),
    ?line GC = lists:keysearch(garbage_collection, 1, B),

    ?line ok.


%% Dummies.

start_spawner() ->
    ok.

stop_spawner() ->
    ok.

bump_reductions(doc) ->
    "Tests erlang:bump_reductions/1.";
bump_reductions(suite) ->
    [];
bump_reductions(Config) when list(Config) ->
    ?line A = 1,
    ?line erlang:garbage_collect(),
    ?line receive after 1 -> ok end,		% Clear reductions.
    ?line {reductions, R1} = process_info(self(), reductions),
    ?line true = erlang:bump_reductions(100),
    ?line B = A + 1,
    ?line {reductions, R2} = process_info(self(), reductions),
    ?line case R2-R1 of
	      Diff when Diff < 100 ->
		  ?line ok = io:format("R1 = ~w, R2 = ~w", [R1, R2]),
		  ?line test_server:fail({small_diff, Diff});
	      Diff when Diff > 110 ->
		  ?line ok = io:format("R1 = ~w, R2 = ~w", [R1, R2]),
		  ?line test_server:fail({big_diff, Diff});
	      Diff ->
		  ok
	  end.

low_prio(doc) ->
    "Priority 'low' should be mixed with 'normal' using a factor of"
    "about 8. (OTP-2644)";
low_prio(suite) ->
    [];
low_prio(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line S = spawn_link(?MODULE, prio_server, [0, 0]),
    ?line P1 = spawn_opt(?MODULE, prio_client, [S, normal], [link, {priority,normal}]),
    ?line P2 = spawn_opt(?MODULE, prio_client, [S, low], [link, {priority,low}]),
    ?line timer:sleep(2000),
    ?line exit(P1, kill),
    ?line exit(P2, kill),
    ?line S ! exit,
    ?line receive {'EXIT', S, {A, B}} -> check_prio(A, B) end,
    ok.

check_prio(A, B) ->
    ?line Prop = A/B,
    ?line ok = io:format("Low=~p, High=~p, Prop=~p\n", [A, B, Prop]),

    %% It isn't 1/8, it's more like 0.3, but let's check that
    %% the low-prio processes get some little chance to run at all.
    ?line true = (Prop < 1.0),
    ?line true = (Prop > 1/32).

prio_server(A, B) ->
    receive
	low ->
	    prio_server(A+1, B);
	normal ->
	    prio_server(A, B+1);
	exit ->
	    exit({A, B})
    end.

prio_client(S, Prio) ->
    S ! Prio,
    prio_client(S, Prio).

make_sub_binary(Bin) when binary(Bin) ->
    {_,B} = split_binary(list_to_binary([0,1,3,Bin]), 3),
    B;
make_sub_binary(List) ->
    make_sub_binary(list_to_binary(List)).



yield(doc) ->
    "Tests erlang:yield/1.";
yield(suite) ->
    [];
yield(Config) when list(Config) ->
    ?line A = 1,
    ?line erlang:garbage_collect(),
    ?line receive after 1 -> ok end,		% Clear reductions.
    ?line SC = schedcnt(start),
    ?line {reductions, R1} = process_info(self(), reductions),
    ?line {ok, true} = call_yield(middle),
    ?line true = call_yield(final),
    ?line true = call_yield(),
    ?line true = apply(erlang, yield, []),
    ?line B = A + 1,
    ?line {reductions, R2} = process_info(self(), reductions),
    ?line Schedcnt = schedcnt(stop, SC),
    ?line case {R2-R1, Schedcnt} of
	      {Diff, 4} when Diff < 30 ->
		  ?line ok = io:format("R1 = ~w, R2 = ~w, Schedcnt = ~w", 
				       [R1, R2, Schedcnt]);
	      {Diff, _} ->
		  ?line ok = io:format("R1 = ~w, R2 = ~w, Schedcnt = ~w", 
				       [R1, R2, Schedcnt]),
		  ?line test_server:fail({measurement_error, Diff, Schedcnt})
	  end.

call_yield() ->
    erlang:yield().

call_yield(middle) ->
    {ok, erlang:yield()};
call_yield(final) ->
    case self() of
	Self when pid(Self) ->
	    ok
    end,
    erlang:yield().

schedcnt(start) ->
    Ref = make_ref(),
    Fun = 
	fun (F, Cnt) ->
		receive
		    {Ref, Parent} ->
			Parent ! {Ref, Cnt}
		after 0 ->
			erlang:yield(),
			F(F, Cnt+1)
		end
	end,
    Pid = spawn_link(fun () -> Fun(Fun, 0) end),
    {Ref, Pid}.

schedcnt(stop, {Ref, Pid}) when reference(Ref), pid(Pid) ->
    Pid ! {Ref, self()},
    receive
	{Ref, Cnt} ->
	    Cnt
    end.
